import java.io.*;
import java.util.*;

/**
 * a utility that manipulates a text file
 * @author Danny Shacham and Yehoyariv Louck
 */
public class CorpusHandler {
	private final int ORIGINAL = 0;
	private final int REVERSE = 1;
	private String delim;
	private Object[] corpus;
	private Trie trie;
	private String fileName;
	private SuffixCounter[] kMostFrequent;
	private int pSimilarity;
	private int k;
	private int order;
	
	/**
	 * constructs a new CorpusHandler with the given "magic" numbers 
	 * @param fileName
	 * @param pSimilarity
	 * @param k
	 * @param delim
	 * @param order
	 */
	public CorpusHandler(String fileName, int pSimilarity, int k, String delim, int order) {
		this.k = k;
		this.pSimilarity = pSimilarity;
		this.order = order;
		this.fileName = fileName;
		this.delim = delim;
		System.out.println("    -- building the Trie");
		trie = getTrie(order);
		System.out.println("    -- finding the "+k+" most frequent suffixes");
		kMostFrequent = generateKMostFrequentSuffixes();
	}
	/**
	 * gets all the corpuse as an array of Strings
	 * @return Object[]
	 */
	public Object[] getCorpusAsArray() {
		return corpus;		
	}
	/** 
	 * gets the k-most frequent suffixes of a corpus
	 * @return SuffixCounter[]
	 */
	public SuffixCounter[] getKMostFrequentSuffixes() {
		return this.kMostFrequent;
	}
	/**
	 * returns the k used in this util to count the k-most-frequent
	 * @return int
	 */
	public int getK() {
		return k;
	}
	/**
	 * changes the p-Similarity that is calculated in this NLP util
	 * @param p the new p-similarity
	 */
	public void changePSimilarity(int p) {
		this.pSimilarity = p;
	}
	/**
	 * changes the k that represents the most frequent suffixes
	 * @param k the k most frequent suffixes
	 */
	public void changeK(int k) {
		this.k = k;
	}
	/**
	 * returns the k most frequent suffixes out of the vector containg the
	 * suffixes
	 * @param suffixes all the suffixes in the Trie
	 * @param k the number of the most frequent suffixes
	 * @return Vector holds the k most frequent suffixes
	 */
	private SuffixCounter[] generateKMostFrequentSuffixes() {
		// gets a suffix as a String
		Vector suffixes = getExplicitSuffix(trie, pSimilarity);
		// divides each suffix to all ot its subStrings
		suffixes = expandSuffix(suffixes);		
		//gets all the suffixes in the corpus
		SuffixCounter[] result = (SuffixCounter[])getKMostFrequentSuffixes(suffixes);
		kMostFrequent = result;
		return result;
	}
	/**
	 * returns the p-similarity of this nlp util
	 * @return int
	 */
	public int getPSimilarity() {
		return pSimilarity;
	}
	/************   private methods    *************************/
	
	/**
	 * gets a Vector containg suffixes as Words and returns an array of
	 * SuffixCounts. each SuffixCounts holds the suffixes from the Vector and
	 * the number of its appearances
	 * @param v the Vector containg all the words, some with multiple
	 * appearances
	 * @return WordCounts[] an array of minimized data
	 */
	private SuffixCounter[] getKMostFrequentSuffixes(Vector v) {
		// Vector of WordCounts
		Vector words = new Vector();
		//Vector of Words
		Vector temp = new Vector();
		for (int i=0; i< v.size(); i++) {
			Word word = (Word)v.elementAt(i);
			int index = temp.indexOf(word);
			if (index == -1) {
				// adds new suffix to the suffix counter
				SuffixCounter w = new SuffixCounter(word);
				words.add(w);
				temp.add(word);
			}
			else {
				//adds this stem to the suffix counter
				SuffixCounter w = (SuffixCounter)words.remove(index);
				w.addSuffix(word);
				words.add(index, w);
			}
		}
		SuffixCounter[] result = new SuffixCounter[k];
		Object[] tempArray = words.toArray();
		Arrays.sort(tempArray);
		System.arraycopy(tempArray, tempArray.length - k, result, 0, k);
		return result;
	}

	/**
	 * gets the explicit suffixes from a given Trie
	 * only those who are being held as a word in the brancehs
	 * @param trie
	 * @param pSimilarity
	 * @return Vector holds all the explicit suffixes of this trie
	 */
	private Vector getExplicitSuffix(Trie trie, int pSimilarity) {
		Vector result = new Vector();
		if (pSimilarity == 0) {
			return trie.getSuffixes();
		}
		Vector v = trie.getBranches();
		// goes recursivly over all the branches looking for similarity of p-1
		for (int i=0; i< v.size(); i++) {
			combineSuffixes(result, getExplicitSuffix((Trie)v.elementAt(i), pSimilarity-1));
		}
		return result;
	}
	/**
	 * gets a Vector of suffixes and expand it that it would also hold all the
	 * suffixes derived from a given suffix.
	 * i.e. if we have a given suffix "ing" it should also have "ng" and "g"
	 * @param suffixes the explicit suffixes
	 * @return Vector all the suffixes and their derivations
	 */
	private Vector expandSuffix(Vector suffixes) {
		Vector result = new Vector();
		for (int i=0; i< suffixes.size(); i++) {
			Word word = (Word)suffixes.elementAt(i);
			String suffix = word.getSuffix();
			String stem = word.getStem();
			// 	goes over all the suffix and adds new suffixes that are
			// a sub string of this one
			for (int j=0; j<= suffix.length(); j++) {
				String suf = suffix.substring(j);
				String newStem = stem+suffix.substring(0,j);
				Word newWord = new Word(newStem, suf);
				result.add(newWord);
			}
		}
		return result;
	}
	
	/**
	 * takes 2 Vectors and makes one Vector with no Multiple data
	 * @param result the original singularity Vector of suffixes
	 * @param words the Vector to be added to result Vector
	 * @return Vector the new combined Vector that holds singularity data
	 */
	private Vector combineSuffixes(Vector result, Vector words) {
		for (int i=0; i<words.size(); i++) {
			Word word = (Word)words.elementAt(i);
			if (!result.contains(word)) {
				result.add(word);
			}
		}
		return result;
	}
	/**
	 * builds a Trie that represents all the data in the given text file
	 * each word is recognized according to the [delim] that is supllied by the
	 * user
	 * according to the order it can build a Trie that holds all the reversed
	 * words
	 * @param order 0 is the original direction 
	 * 1 is reversed direction
	 * @return Trie
	 */
	private  Trie getTrie(int order) {
		Vector temp = new Vector();
		Trie trie = new Trie();
		try {			
			BufferedReader in = new BufferedReader(new InputStreamReader(new FileInputStream(fileName)));
			String nextLine = in.readLine();
			while (nextLine != null ){
				 StringTokenizer line = new StringTokenizer(nextLine, delim);
				 while (line.hasMoreTokens()) {
				 	String word = line.nextToken();
				 	word = word.toLowerCase();
				 	if (order == REVERSE) {
			 			StringBuffer  sb = new StringBuffer(word);
			 			word = sb.reverse().toString();
				 	}
				 	trie.put(word);
				 	temp.add(word);
				 }
				nextLine = in.readLine();
			}
		} catch (FileNotFoundException fnfe) {}
		catch (IOException ioe) {}
		corpus = temp.toArray();
		return trie;
	}			 	


}
