import jmat.data.*;
//import jmat.data.matrixDecompositions.*;
import java.util.*;

/**
 * LSAHandler is a util class that handles a corpus and computes the M Matrix
 * using latent semantic vectors (LSA)
 * the values of this Matrix are z-scores rather than counts
 * @author Danny Shacham and Yehoyariv Louck
 *
 */
public class LSAHandler {

	private Matrix M;
	// k dimantional V matrix
	private double[][] Vk;	
	// mapping the words that are not in the N most frequent word 
	// to their own Matrix row
	private Hashtable restOfWords, NWords;
	private int N, vk;
	private int random;
	// the window that we build the M matrix with
	private int interval;
	private int[] randomVectorsIndex;
	// an array of the most frequent words in the corpus
	private WordCounter[] mostFrequent;
	private Object[] corpus;
	private Matrix randomMatrix;
	/**
	 * constructs a new LSAHandler
	 * @param fileName the corpus file name
	 * @param N the number of the most frequent words in this corpus
	 */
	public LSAHandler(int N, int random, int vk, Object[] corpus, int interval) {
		this.N = N;
		this.random = random;
		this.corpus = corpus;
		this.interval = interval;
		M = null;
		this.vk = vk;
		restOfWords = new Hashtable();
		NWords = new Hashtable();
		System.out.println(" - finding the "+N+" most frequent words in the corpora");
		mostFrequent = generateNWords();
		randomVectorsIndex = generateRandom(random);
		System.out.println(" - building the M matrix");
		buildMatrix();	
		generateRandomMatrix();		
	}
	/**
	 * computes the M Matrix as expleind at Schone & Jurafsky'w article
	 * @return Matrix 
	 */
	public Matrix getMatrix() {
		return M;		
	}
	/**
	 * gets the N most frequent words in the corpus
	 * @return WordCounter[] an array holding the N most frequent words
	 */
	public WordCounter[] getNMostFrequentWords() {
		return this.mostFrequent;
	}
	/**
	 * computing the NCS ( normalized cosine score ) between 2 words
	 * @return double the NCS ofthe 2 words
	 */
	public double computeNCS(PPMV ppmv) {
		Word word1 = ppmv.getWord1();
		Word word2 = ppmv.getWord2();
		Matrix omegaW = new Matrix(1, randomVectorsIndex.length);
		Matrix omegaV = new Matrix(1, randomVectorsIndex.length);
		Matrix w = getVector(word1.getStem()+word1.getSuffix());
		Matrix v = getVector(word2.getStem()+word2.getSuffix());
		// computing the cosine between each word to RANDOM_VECTORS
		// random words from the top N most frequent words
		for (int i=0; i< randomVectorsIndex.length; i++) {
			Matrix temp = randomMatrix.getRow(i);
			omegaW.set(0, i, computeCosine(w, temp));
			omegaV.set(0, i, computeCosine(v, temp));			
		}
		omegaW = omegaW.transpose();
		omegaV = omegaV.transpose();
		// the correlation mean
		double mW = omegaW.mean().get(0,0);
		double mV = omegaW.mean().get(0,0);
		// the correlation standarddeviation
		double sW = Math.sqrt(omegaW.variance().get(0,0));
		double sV = Math.sqrt(omegaW.variance().get(0,0));
		double temp = computeCosine(w,v);
		double arg1 = (temp-mW)/sW;
		double arg2 = (temp-mV)/sV;
		double result = Math.min(arg1, arg2);
		return result;
	}		
		 
	/* *****************		private methods		***************** */
	/**
	 * compute the cosine between the vectors of 2 words
	 */
	private double computeCosine(Matrix v1, Matrix v2) {
		double mul = v1.times(v2.transpose()).get(0,0);
		v1 = ((v1.ebePow(2)).transpose()).sum();
		v2 = ((v2.ebePow(2)).transpose()).sum();		 
		double cos = mul/(v1.get(0,0)*v2.get(0,0));
		return cos;
	}
	/**
	 * generates a matrix of [random] rows
	 * each row is a Vector of a random word in the corpora
	 */	
	private void generateRandomMatrix() {
		randomMatrix = new Matrix(randomVectorsIndex.length, 2*N);
		for (int i=0; i< randomVectorsIndex.length; i++) {
			Matrix temp = getVector(mostFrequent[randomVectorsIndex[i]].getWord().getSuffix());
			randomMatrix.setRow(i, temp);
		}
	}
	/**
	 * gets the vector of this word
	 * @param word
	 * @return Matrix
	 */
	private Matrix getVector(String word) {
		int indexWord1 = getIndexOf(word);
		Matrix vector;
		if (indexWord1>=0) {
			vector = M.getRow(indexWord1);
		} else {
			vector = (Matrix)restOfWords.get(word);
		}
		Matrix temp = new Matrix(Vk);
		//vector = temp.p 
		return vector;
	}
	/**
	 * builds the M matrix, and saves the wector for those words that are not
	 * int he top N most frequent words in a Hashtable
	 */
	private void  buildMatrix() {
		M = new Matrix(N, 2*N);
		Matrix temp;
		int size = corpus.length;
		System.out.println("    -- filling the matrix");
		// computing the right side words
		System.out.println("       -- the corpora size is "+size+" words");
		System.out.println("          the progress is:");
		System.out.print  ("          -- ");
		int counter = 0;
		for (int i=0; i< size-interval; i++) {
			if (i%(size/10) == 0) {
				System.out.print(counter+"/10"+"  ");
				counter++;
			}
			String first = (String)corpus[i];
			for (int j=1; j< interval; j++) {
				String second = (String)corpus[i+j];
				int indexFirst = getIndexOf(first);
				int indexSecond = getIndexOf(second);
				// if the word is not in the top N most frequent words it put 
				// the values in the Nth row
				if (indexFirst < 0) {
					indexFirst = N-1;
				}
				if (indexSecond < 0) {					
					indexSecond = N-1;
				}
				// each right side word first is also a left side 
				// word of second
				double temp1 = M.get(indexFirst, N+indexSecond);
				double temp2 = M.get(indexSecond, indexFirst);
				M.set(indexFirst, N+indexSecond, temp1+1);
				M.set(indexSecond, indexFirst, temp2+1);
				// if the first word is not in the N most frequent words
				// it puts the values also in a seperate vector
				if (indexFirst == N-1) {
					temp = (Matrix)restOfWords.get(first);
					if (temp == null) {
						temp = new Matrix(1,2*N);
					}
					temp1 = M.get(indexFirst, N+indexSecond);
					temp.set(0, N+indexSecond, temp1+1);
					restOfWords.put(first, temp);
				}
				if (indexSecond == N-1) {
					temp = (Matrix)restOfWords.get(second);
					if (temp == null) {
						temp = new Matrix(1,2*N);
					}
					temp2 = M.get(indexSecond, indexFirst);
					temp.set(0, indexFirst, temp2+1);
					restOfWords.put(second,temp);
				}				
			}					
		}
		System.out.println();
		System.out.println("    -- getting the V matrix");	
		Vk = new double[1][vk];
		//SingularValueDecomposition svd = M.svd();
		//double[] tempArray = svd.getSingularValues();
		//System.out.println("       -- the length of the Vk matrix is: "+tempArray.length);
		//System.arraycopy(tempArray, 0, Vk[0], 0, Vk[0].length); 
	}
	/**
	 * generates an array of size [size] each item is a random ranber
	 * @param size the size of the array
	 * @return int[]
	 */
	private int[] generateRandom(int size) {
		int[] result = new int[size];
		for (int i=0; i< size; i++) {
			int temp = (int)Math.round(Math.random()*(double)(N-2));
			result[i] = temp;			
		}
		return result;
	}
			
	/**
	 * getting the index of a word in the mostFrequent array
	 * @param word
	 * @return int -1 if the word is not in the N most Frequent words
	 */
	private int getIndexOf(String word) {
		int result = -1;
		Object temp = NWords.get(word);
		if (temp == null) {
			return result;
		}
		result = ((Integer)temp).intValue();
		return result;
	}
	/**
	 * finds the top N most frequent words
	 * @return WordCounter[] an array of the top N most frequent words
	 */
	private WordCounter[] generateNWords() {
		Vector allWords = new Vector();
		WordCounter[] result = new WordCounter[N-1];
		for (int i=0; i< corpus.length; i++) {
			String word = (String)corpus[i];
			Word newWord = new Word("", word);
			WordCounter wordCount = new WordCounter(newWord);
			int index = allWords.indexOf(wordCount);
			if (index >=0) {
				((WordCounter)allWords.elementAt(index)).addWord();
			}else {
				allWords.add(wordCount);
			}					
		}
		Object[] temp = allWords.toArray();
		Arrays.sort(temp);
		System.arraycopy(temp, temp.length-N-1, result, 0, N-1);
		for (int i= 0; i< result.length; i++) {
			String word = result[i].getWord().getSuffix();
			NWords.put(word, new Integer(i));			
		}		
		return result;
	}
}
