import jmat.data.*;
import java.util.*;
import java.io.*;
/**
 * a NLPUtil calss is a util class that implements the Schone and Jorafsky
 * algorithm as described in "Knowledge-Free Inductionof Morphology Using
 * LatentSemantic Analysis"
 * @author Danny Shacham and Yehoyariv Louck
 *
 */
public class NLPUtil {
	// the direction of the words ( from start to end or vise versa
	private final int ORIGINAL = 0;
	private final int REVERSE = 1;
	
	private String fileName;
	private Matrix M;
	private int N;
	public static SuffixCounter[] kMostFrequentSuffixes;
	public static Object[] corpus;
	private int k, vk;
	private int pSimilarity;
	private int random;
	private int interval;
	private RuleSet[] rules;
	private Object[] NCS;
	private String delim;
	private String lang;
	private String posFile;
	private String negFile;
	
	/**
	 * constructs a new NLPUtil with all the "magic" numbers 
	 */
	public NLPUtil(String fileName, int pSimilarity, int k, int vk, int N, int random, int interval, String delim, String lang, String posFile, String negFile) {
		this.fileName = fileName;
		this.k = k;
		this.vk = vk;
		this.pSimilarity = pSimilarity;
		this.N = N;
		this.random = random;
		this.interval = interval;
		this.delim = delim;
		this.lang = lang;
		this.posFile = posFile;
		this.negFile = negFile;
		System.out.println("1. getting the RuleSets: ");
		rules = generateRuleSets();
		System.out.println("2. computing the NCS: ");
		NCS = generateNCS();
		System.out.println(" - Done!");
	}
	/**
	 * gets all the rulesets of this corpora
	 * @return RuleSet[] an array of rulesets
	 */
	public RuleSet[] getRuleSets() {
		return rules;
	}
	/**
	 * gets the NCS of every PPMV in the rulesets
	 * @return Object[] an array of NCS in the size of all the PPMVs 
	 */
	public Object[] getNCS() {
		return NCS;
	}
	/**
	 * prints the NCS to 2 files
	 * out1 - all the PPMVs above a threshold 
	 * out2 - all the PPMVs below it
	 */
	public void printNCS(int thresh) {
		try {
			PrintWriter out1 = new PrintWriter(new BufferedWriter(new FileWriter(posFile)));
			PrintWriter out2 = new PrintWriter(new BufferedWriter(new FileWriter(negFile)));
			out1.println("**  PPMV   ***               NCS   ***");
			out1.println("---------------------------------------");
			out2.println("**  PPMV   ***               NCS   ***");
			out2.println("---------------------------------------");
			for (int i=0; i< NCS.length; i++) {
				NCS temp = (NCS)NCS[i];
				if (temp.getNCS() > thresh) {
					out1.println(temp);
				} else {
					out2.println(temp);
				}
			}
			out1.flush();
			out1.close();
			out2.flush();
			out2.close();
		}catch (IOException ioe) {
			System.out.println("there were problems writing to the file");
		}
	}
	
	/* **************************     private methods      ************ */
	/**
	 * generates the NCS array using LSA technic
	 */
	private Object[] generateNCS() {
		LSAHandler lsa = new LSAHandler(N, random, vk, corpus, interval);
		System.out.println(" - getting the NCS of each PPMV");
		int size = rules.length;
		int numOfPPMVs = 0;
		for (int i=0; i< size; i++){
			numOfPPMVs += rules[i].getAllPPMV().size();
		}
		System.out.println("   -- the number of PPMVs is: "+numOfPPMVs+" PPMVs");
		System.out.println("      the progress is:");
		System.out.print  ("      -- ");
		Object[] result;
		int index = 0;
		int counter = 0;
		Vector tempResult = new Vector();
		// going over all the rules
		for (int j=0; j< size; j++) {				
			Vector v = rules[j].getAllPPMV();
			Object[] ppmvs = v.toArray();
			// going over all the PPMVs in a rule
			for (int i=0; i<v.size(); i++) {
				if (index%(numOfPPMVs/10) == 0) {
					System.out.print(counter+"/10"+"  ");
					counter++;
				}
				PPMV p = (PPMV)ppmvs[i];
				// compute the NCS for the 2 words in the PPMV
				double ncs = lsa.computeNCS(p);
				NCS ncsTemp = new NCS(p, ncs);
				tempResult.add(ncsTemp);				
				index++;				
			}
		}
		result = tempResult.toArray();
		System.out.println();
		Arrays.sort(result);			
		return result;	
	}	
	/**
	 * generates the rulesets from a corpora
	 * @return RuleSet[]
	 */
	private RuleSet[] generateRuleSets() {
		System.out.println(" - manipulating the corpora");
		CorpusHandler corpusHandler;
		corpusHandler = new CorpusHandler(fileName,pSimilarity, k, delim, 0);
		SuffixCounter[] kMostFrequentAffixes = null;
		// getting the k-most frequent suffixes from the corpora
		// if the language is hebrew then it also gets the k-most frequent affixes
		if (lang.equalsIgnoreCase("english")) {			
			kMostFrequentSuffixes = corpusHandler.getKMostFrequentSuffixes();			
		} else {
			//kMostFrequentAffixes = corpusHandler.getKMostFrequentSuffixes();
			corpusHandler = new CorpusHandler(fileName,pSimilarity, k, delim, 1);
			kMostFrequentSuffixes = corpusHandler.getKMostFrequentSuffixes();
		}
		corpus = corpusHandler.getCorpusAsArray();		
		System.out.println(" - generating the RuleSets");
		// getting the rulesets according to the k-most frequent suffixes 
		// ( and affixes too if Hebrew)
		RuleSet[] rules = PPMVHandler.getRuleSets(kMostFrequentSuffixes, kMostFrequentAffixes);
		return rules;
	}	 

}
