import java.util.*;
/**
 * a PPMVHandler is util a class that can take the k-most frequent suffixes in
 * the corpora and to get all the rule-sets containing all the PPMVs acording to
 * given p-similarity and k
 * @author Danny Shacham and Yehoyariv Louck
 */
public class PPMVHandler {
	
	
	private PPMVHandler() {}

	/**
	 * returns an arry of size k*k that holds all the combinations of rules and
	 * all the PPMVs that hold these rules
	 * @return RuleSet[]
	 */
	public static RuleSet[] getRuleSets(SuffixCounter[] kMostFrequentsuffixes, SuffixCounter[] kMostFrequentAffixes) {
		RuleSet[] result = createEmptyRuleSets(kMostFrequentsuffixes);
		System.out.println("    -- filling the RuleSets with PPMVs");		
		result = fillRuleSets(kMostFrequentsuffixes, kMostFrequentAffixes, result);		
		return result;
	}

	/****************		private methods 		************************/
	/**
	 * fills the empty rule-sets with the words from the corpus. 
	 * @param empty an empty rule-sets array that holds the rules of each set
	 * from hte k*k possible sets
	 * @return RuleSet[] the full rule-sets array
	 */
	private static RuleSet[] fillRuleSets(SuffixCounter[] kMostFrequent, SuffixCounter[] kMostFrequentAffixes, RuleSet[] empty) {
		for (int i=0; i< empty.length; i++){
			SuffixCounter suf1 = getSuffixCounter(kMostFrequent, empty[i].getRule().getSuffix1());
			SuffixCounter suf2 = getSuffixCounter(kMostFrequent, empty[i].getRule().getSuffix2());
			empty[i] = insertPPMVToRuleSet(suf1, suf2, kMostFrequentAffixes);			
		}
		return empty;
	}
	/**
	 * gets the suffixCounter with the suffix "suffix" from the k most frequent
	 * suffixes
	 * @param suffix the suffix we want its counter
	 * @return SuffixCounter the suffix counter that has a suffix "suffix"
	 */
	private static SuffixCounter getSuffixCounter(SuffixCounter[] kMostFrequent, String suffix) {
		for (int i=0; i< kMostFrequent.length; i++) {			
			if (kMostFrequent[i].equals(new SuffixCounter(new Word("",suffix)))) {
				return kMostFrequent[i];
			}			
		}
		return null;
	}
	/**
	 * insert all the PPMVs that hold a certain rule to the rule set
	 * @param suf1
	 * @param suf2
	 * @return RuleSet
	 */		
	private static RuleSet insertPPMVToRuleSet(SuffixCounter suf1, SuffixCounter suf2, SuffixCounter[] kMostFrequentAffixes) {
		Rule rule = new Rule(suf1.getSuffix(), suf2.getSuffix());
		RuleSet result = new RuleSet(rule);
		if (suf1 != null && suf2 != null) {
			Vector stems1 = suf1.getStems();
			Vector stems2 = suf2.getStems();
			// goes over all the stems that have suf1 and suf2
			// it the stems are equale it adds it to the ruleSet as PPMV
			for (int i=0; i<stems1.size(); i++) {
				String stem = (String)stems1.elementAt(i);
				for (int j=0; j<stems2.size(); j++) {
					if (kMostFrequentAffixes == null) {
						if (stem.equals(stems2.elementAt(j))) {
							result.addPPMV(new PPMV(rule, stem));
						}												
					} 
				}
			}
		}
		return result;		
	}
	
	/**
	 * creates all the rule-sets, with no PPMVs in them
	 * the rule-sets are every combination of 2 rules, according to the k most
	 * frequent suffixes
	 * @return RuleSet[] empty rule-sets each one with a specific rule
	 */
	private static RuleSet[] createEmptyRuleSets(SuffixCounter[] kMostFrequent) {
		int k  = kMostFrequent.length;
		// all the combinations of choosing couples out of k elements
		int numOfCombinations = k*(k-1)/2;
		RuleSet[] result = new RuleSet[numOfCombinations];
		int m=0;
		for (int i=0; i<k-1; i++){
			for (int j=i+1; j<k; j++){
				// creates a rule out of 2 suffixes
				Rule rule = new Rule(kMostFrequent[i].getSuffix(), kMostFrequent[j].getSuffix());
				result[m] = new RuleSet(rule);
				m++;
			}
		}	
		return result;
	}
}
